""" Compute heads of mentions. """

import re

from cort.core import spans

__author__ = 'smartschat'


class HeadFinder:
    """Compute heads of mentions.

    This class provides functions to compute heads of mentions via modified
    version of the rules that can be found in Michael Collins' PhD thesis.
    The following changes were introduced:

        - handle NML as NP,
        - for coordinated phrases, take the coordination token as head,

    Furthermore, this class provides a function for adjusting heads for proper
    names to multi-token phrases via heuristics (see adjust_head_for_nam).
    """
    def __init__(self):
        self.__nonterminals = ["NP", "NML", "VP", "ADJP", "QP", "WHADVP", "S",
                             "ADVP", "WHNP", "SBAR", "SBARQ", "PP", "INTJ",
                             "SQ", "UCP", "X", "FRAG"]


        self.__nonterminal_rules = {
            "VP": (["TO", "VBD", "VBN", "MD", "VBZ", "VB", "VBG", "VBP", "VP",
                     "ADJP", "NN", "NNS", "NP"], False),
            "ADJP": (["NNS", "QP", "NN", "\$", "ADVP", "JJ", "VBN", "VBG", "ADJP",
              "JJR", "NP", "JJS", "DT", "FW", "RBR", "RBS", "SBAR", "RB"],
                     False),
            "QP": (["\$", "NNS", "NN", "IN", "JJ", "RB", "DT", "CD", "NCD",
              "QP", "JJR", "JJS"], False),
            "WHADVP": (["CC", "WRB"], True),
            "S": (["TO", "IN", "VP", "S", "SBAR", "ADJP", "UCP", "NP"], False),
            "SBAR": (["WHNP", "WHPP", "WHADVP", "WHADJP", "IN", "DT", "S", "SQ",
              "SINV", "SBAR", "FRAG"], False),
            "SBARQ": (["SQ", "S", "SINV", "SBARQ", "FRAG"], False),
            "SQ": (["VBZ", "VBD", "VBP", "VB", "MD", "VP", "SQ"], False),
            "ADVP": (["RB", "RBR", "RBS", "FW", "ADVP", "TO", "CD", "JJR", "JJ",
              "IN", "NP", "JJS", "NN"], True),
            "WHNP": (["WDT", "WP", "WP$", "WHADJP", "WHPP", "WHNP"], True),
            "PP": (["IN", "TO", "VBG", "VBN", "RP", "FW"], True),
            "X": (["S", "VP", "ADJP", "JJP", "NP", "SBAR", "PP", "X"], True),
            "FRAG": (["*"], True),
            "INTJ": (["*"], False),
            "UCP": (["*"], True),
        }

    def get_head(self, tree):
        """
        Compute the head of a mention, which is represented by its parse tree.

        Args:
            tree (nltk.ParentedTree): The parse tree of a mention.

        Returns:
            nltk.ParentedTree: The subtree of the input tree which corresponds
            to the head of the mention.
        """
        head = None

        label = tree.label()

        if len(tree) == 1:
            if tree.height() == 3:
                head = tree[0]
            elif tree.height() == 2:
                head = tree
        elif label in ["NP", "NML"]:
            head = self.__get_head_for_np(tree)
        elif label in self.__nonterminals:
            head = self.__get_head_for_nonterminal(tree)

        if head is None:
            head = self.get_head(tree[-1])

        return head

    def __get_head_for_np(self, tree):
        if self.__rule_cc(tree) is not None:
            return self.__rule_cc(tree)
        elif self.__collins_rule_nn(tree) is not None:
            return self.__collins_rule_nn(tree)
        elif self.__collins_rule_np(tree) is not None:
            return self.get_head(self.__collins_rule_np(tree))
        elif self.__collins_rule_nml(tree) is not None:
            return self.get_head(self.__collins_rule_nml(tree))
        elif self.__collins_rule_prn(tree) is not None:
            return self.__collins_rule_prn(tree)
        elif self.__collins_rule_cd(tree) is not None:
            return self.__collins_rule_cd(tree)
        elif self.__collins_rule_jj(tree) is not None:
            return self.__collins_rule_jj(tree)
        elif self.__collins_rule_last_word(tree) is not None:
            return self.__collins_rule_last_word(tree)

    def __get_head_for_nonterminal(self, tree):
        label = tree.label()
        values, traverse_reversed = self.__nonterminal_rules[label]
        if traverse_reversed:
            to_traverse = reversed(tree)
        else:
            to_traverse = tree
        for val in values:
            for child in to_traverse:
                label = child.label()
                if val == "*" or label == val:
                    if label in self.__nonterminals:
                        return self.get_head(child)
                    else:
                        return child

    def __rule_cc(self, tree):
        if tree.label() == "NP":
            for child in tree:
                if child.label() == "CC":
                    return child

    def __collins_rule_pos(self, tree):
        if tree.pos()[-1][1] == "POS":
            return tree[-1]

    def __collins_rule_nn(self, tree):
        for i in range(len(tree)-1, -1, -1):
            if re.match("NN|NNP|NNPS|JJR", tree[i].label()):
                return tree[i]
            elif tree[i].label() == "NX":
                return self.get_head(tree[i])

    def __collins_rule_np(self, tree):
        for child in tree:
            if child.label() == "NP":
                return child

    def __collins_rule_nml(self, tree):
        for child in tree:
            if child.label() == "NML":
                return child

    def __collins_rule_prn(self, tree):
        for child in tree:
            if child.label() == "PRN":
                return self.get_head(child[0])

    def __collins_rule_cd(self, tree):
        for i in range(len(tree)-1, -1, -1):
            if re.match("CD", tree[i].label()):
                return tree[i]

    def __collins_rule_jj(self, tree):
        for i in range(len(tree)-1, -1, -1):
            if re.match("JJ|JJS|RB", tree[i].label()):
                return tree[i]
            elif tree[i].label() == "QP":
                return self.get_head(tree[i])

    def __collins_rule_last_word(self, tree):
        current_tree = tree[-1]
        while current_tree.height() > 2:
            current_tree = current_tree[-1]

    @staticmethod
    def adjust_head_for_nam(tokens, pos, ner_type):
        """
        Adjust head for proper names via heuristics.

        Based on heuristics depending on the named entity type (person,
        organization, ...) and part-of-speech tags, adjust the head of a
        named entity mention to a meaningful extent useful for coreference
        resolution.

        For example, for the mention "Khan Younes in Southern Gaza Strip",
        this function will compute "Khan Younes" as the head.

        Args:
            tokens (list(str)): The tokens of the mention.
            pos (list(str)): The part-of-speech tags of the mention.
            ner_type (str): The named entity type of the mention. Should be
                one of PERSON, ORG, GPE, FAC, NORP, PRODUCT, EVENT, MONEY,
                WORK_OF_ART, LOC, LAW, LANGUAGE, DATE, TIME, ORDINAL,
                CARDINAL, QUANTITY, PERCENT or NONE.

        Returns:
            (Span, list(str)):
        """
        # TODO: get rid of this ugly hack
        if len(pos) == 0:
            return spans.Span(0, 0), "NOHEAD"

        stop_regex = re.compile("CC|,|\.|:|;|V.*|IN|W.*|ADVP|NN$")

        if re.match("ORG.*|GPE.*|FAC.*|NORP.*|PRODUCT|EVENT|MONEY|" +
                    "WORK_OF_ART|LOC.*|LAW|LANGUAGE", ner_type):
            start_regex = re.compile("NN(S)?|NNP(S)?")
            stop_regex = re.compile("V.*|IN|W.*|ADVP|,|-LRB-")
        elif ner_type == "PERSON":
            start_regex = re.compile("NN(S)?|NNP(S)?")
            stop_regex = re.compile("IN|CC|,|\.|:|;|V.*|W.*|-LRB-")
        elif re.match("DATE|TIME", ner_type):
            start_regex = re.compile("NN(S)?|NNP(S)?|CD")
        elif re.match("ORDINAL", ner_type):
            start_regex = re.compile("NN|JJ|RB")
        elif re.match("CARDINAL", ner_type):
            start_regex = re.compile("CD")
        elif re.match("QUANTITY|PERCENT", ner_type):
            start_regex = re.compile("CD|JJ|NN")
        elif ner_type == "NONE":
            start_regex = re.compile("NN(S)?|NNP(S)?|CD")
        else:
            raise Exception("Unknown named entity annotation: " + ner_type)

        head_start = -1

        position = 0

        for i in range(0, len(tokens)):
            position = i
            if head_start == -1 and start_regex.match(pos[i]):
                head_start = i
            elif head_start >= 0 and stop_regex.match(pos[i]):
                return spans.Span(head_start, i-1), tokens[head_start:i]

        if head_start == -1:
            head_start = 0

        if pos[position] == "POS" and position == len(pos) - 1:
            position -= 1

        return spans.Span(head_start, position), tokens[head_start:position+1]



import unittest

from cort.analysis import data_structures
from cort.analysis import spanning_tree_algorithms
from cort.core import mentions
from cort.core import spans


__author__ = 'smartschat'


class TestSpanningTreeAlgorithms(unittest.TestCase):
    def setUp(self):
        self.gold_first_cluster = [
            mentions.Mention(
                None,
                spans.Span(0, 0),
                {"tokens": ["a"], "type": "NOM", "annotated_set_id": 0}),

            mentions.Mention(
                None,
                spans.Span(1, 1),
                {"tokens": ["US"], "type": "NAM", "annotated_set_id": 0}),

            mentions.Mention(
                None,
                spans.Span(2, 3),
                {"tokens": ["angry", "salesman"], "type": "PRO", "annotated_set_id": 0}),

            mentions.Mention(
                None,
                spans.Span(4, 5),
                {"tokens": ["the", "rainbow"], "type": "NAM",
                 "annotated_set_id": 0}),

            mentions.Mention(
                None,
                spans.Span(5, 6),
                {"tokens": ["and", "far"], "type": "NOM",
                 "annotated_set_id": 0}),

            mentions.Mention(
                None,
                spans.Span(7, 7),
                {"tokens": ["neypmd"], "type": "NOM", "annotated_set_id": 0}),
        ]

        self.gold_second_cluster = [
            mentions.Mention(
                None,
                spans.Span(7, 8),
                {"type": "NOM", "annotated_set_id": 1}),

            mentions.Mention(
                None,
                spans.Span(9, 9),
                {"type": "NAM", "annotated_set_id": 1}),

            mentions.Mention(
                None,
                spans.Span(10, 10),
                {"type": "PRO", "annotated_set_id": 1}),
        ]

        self.system1_mentions = [
            mentions.Mention(None, spans.Span(0, 0), {"set_id": 0}),
            mentions.Mention(None, spans.Span(2, 3), {"set_id": 0}),
            mentions.Mention(None, spans.Span(4, 5), {"set_id": 2}),
            mentions.Mention(None, spans.Span(5, 6), {"set_id": 2}),
            mentions.Mention(None, spans.Span(3, 4), {"set_id": 1}),
            mentions.Mention(None, spans.Span(7, 8), {"set_id": 1}),
        ]

        self.system2_cluster = [
            mentions.Mention(
                None,
                spans.Span(0, 0),
                {"tokens": ["a"], "set_id": 0}),

            mentions.Mention(
                None,
                spans.Span(2, 3),
                {"tokens": ["angry", "salesman"], "set_id": 0}),

            mentions.Mention(
                None,
                spans.Span(7, 8),
                {"tokens": ["snafu", "foo"], "set_id": 0}),

            mentions.Mention(
                None,
                spans.Span(9, 9),
                {"tokens": ["bar"], "set_id": 0}),
        ]
        self.system2_cluster[1].attributes["antecedent"] = \
            self.system2_cluster[0]
        self.system2_cluster[2].attributes["antecedent"] = \
            self.system2_cluster[0]
        self.system2_cluster[3].attributes["antecedent"] = \
            self.system2_cluster[2]

        self.maxDiff = None

    def test_recall_closest(self):
        gold_graph = data_structures.EntityGraph.from_mentions(
            self.gold_first_cluster, "annotated_set_id")[0]

        spanning_tree_edges = [
            (self.gold_first_cluster[1], self.gold_first_cluster[0]),
            (self.gold_first_cluster[2], self.gold_first_cluster[0]),
            (self.gold_first_cluster[3], self.gold_first_cluster[2]),
            (self.gold_first_cluster[4], self.gold_first_cluster[3]),
            (self.gold_first_cluster[5], self.gold_first_cluster[4])
        ]

        self.assertEqual(
            spanning_tree_edges,
            spanning_tree_algorithms.recall_closest(
                gold_graph,
                gold_graph.partition(
                    data_structures.EntityGraph.from_mentions(
                        self.system1_mentions, "set_id"))))

    def test_recall_type(self):
        gold_graph = data_structures.EntityGraph.from_mentions(
            self.gold_first_cluster, "annotated_set_id")[0]

        spanning_tree_edges = [
            (self.gold_first_cluster[1], self.gold_first_cluster[0]),
            (self.gold_first_cluster[2], self.gold_first_cluster[0]),
            (self.gold_first_cluster[3], self.gold_first_cluster[1]),
            (self.gold_first_cluster[4], self.gold_first_cluster[3]),
            (self.gold_first_cluster[5], self.gold_first_cluster[3])
        ]

        self.assertEqual(
            spanning_tree_edges,
            spanning_tree_algorithms.recall_accessibility(
                gold_graph,
                gold_graph.partition(
                    data_structures.EntityGraph.from_mentions(
                        self.system1_mentions, "set_id"))))

    def test_precision_system_output(self):
        gold_graph = data_structures.EntityGraph.from_mentions(
            self.system2_cluster, "set_id")[0]

        spanning_tree_edges = [
            (self.system2_cluster[1], self.system2_cluster[0]),
            (self.system2_cluster[2], self.system2_cluster[0]),
            (self.system2_cluster[3], self.system2_cluster[2])
        ]

        self.assertEqual(
            spanning_tree_edges,
            spanning_tree_algorithms.precision_system_output(
                gold_graph,
                gold_graph.partition(
                    data_structures.EntityGraph.from_mentions(
                        self.gold_first_cluster, "annotated_set_id"))))


if __name__ == '__main__':
    unittest.main()



from distutils.core import setup


setup(
    name='cort',
    version='0.2.4.2',
    packages=['cort',
              'cort.analysis',
              'cort.core',
              'cort.test',
              'cort.coreference',
              'cort.test.multigraph',
              'cort.test.analysis',
              'cort.test.core',
              'cort.coreference.multigraph',
              'cort.coreference.approaches',
              'cort.util',
              'cort.preprocessing',
              'stanford_corenlp_pywrapper'],

    url='http://github.com/smartschat/cort',
    license='MIT',
    author='Sebastian Martschat, Thierry Goeckel, Patrick Claus',
    author_email='sebastian.martschat@gmail.com',
    description='A coreference resolution research toolkit.',
    keywords = ['NLP', 'CL', 'natural language processing',
                'computational linguistics', 'coreference resolution',
                'text analytics'],
    classifiers = [
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Text Processing',
        ],
    install_requires=['nltk >= 3.0.1', 'numpy', 'matplotlib', 'mmh3', 'cython',
                      'future', 'jpype1', 'beautifulsoup4',
                      'pystanforddependencies >= 0.2.0'],
    package_data={
        'cort': ['analysis/visualization/style.css',
                 'analysis/visualization/lib/*',
                 'resources/*',
                 'config_files/*',
                 'coreference/perceptrons.pyx',
                 "reference-coreference-scorers/v8.01/*.*",
                 "reference-coreference-scorers/v8.01/lib/*.pm",
                 "reference-coreference-scorers/v8.01/lib/Algorithm/*",
                 "reference-coreference-scorers/v8.01/lib/Data/*",
                 "reference-coreference-scorers/v8.01/lib/Math/*"],
        'stanford_corenlp_pywrapper': ['rcorenlp.r',
                                       'lib/*',
                                       'javasrc/corenlp/*',
                                       'javasrc/util/misc/*',
                                       'javasrc/util/*.java'],
    },
    scripts=['bin/cort-train', 'bin/cort-predict-conll',
             'bin/cort-predict-raw', 'bin/cort-visualize',
             'bin/run-multigraph']
)



""" Implements instance extraction and decoding for mention-ranking models.

This module implements two variants of the mention ranking model within a
framework that expresses coreference resolution as predicting latent structures,
while performing learning using a latent structured perceptron with
cost-augmented inference.

Hence, both variants are expressed as predicting a latent graph. In particular,
let m_1, ..., m_n be all mentions in a document. Let m_0 be a dummy mention for
anaphoricity determination. For an anaphor m_j, we predict the graph with nodes
m_0, ..., m_j and with the arc (m_j, m_i) if m_i is selected as the antecedent
of m_j. Such a graph is called a *substructure* (compared to the graph which
summarizes the decisions for the whole document).

The two variants implemented here are

    - latent antecedents for training (Chang et al., 2012): when learning
      weights, compare the prediction (m_j,m_i) with the best-scoring
      prediction (m_j,m_k) s.t. m_j and m_k are coreferent
    - closest antecedents for training (Denis and Baldridge, 2008): when
      learning weights, compare the prediction (m_j,m_i) with (m_j,m_k), where
      m_k is the closest antecedent of m_j.

To implement these variants, this module contains a function that defines the
search space for the graphs, and two decoders: one decoder computes the
best-scoring antecedent prediction and the best-scoring coreferent antecedent,
while the other computes the best-scoring antecedent prediction and the closest
antecedent.

References:

    - Pascal Denis and Jason Baldridge. 2008. Specialized models and ranking
      for coreference resolution. In *Proceedings of the 2008 Conference on
      Empirical Methods in Natural Language Processing*, Waikiki, Honolulu,
      Hawaii, 25-27 October 2008, pages 660-669.
      http://www.aclweb.org/anthology/D08-1069
    - Kai-Wei Chang, Rajhans Samdani, Alla Rozovskaya, Mark Sammons, and
      Dan Roth. 2012. Illinois-Coref: The UI system in the CoNLL-2012 shared
      task. In *Proceedings of the Shared Task of the 16th Conference on
      Computational Natural Language Learning*, Jeju Island, Korea, 12-14 July
      2012, pages 113-117.
      http://www.aclweb.org/anthology/W12-4513
"""


from cort.coreference import perceptrons


__author__ = 'martscsn'


def extract_substructures(doc):
    """ Extract the search space for the mention ranking model,

    The mention ranking model consists in computing the optimal antecedent
    for an anaphor, which corresponds to predicting an edge in graph. This
    functions extracts the search space for each such substructure (one
    substructure corresponds to one antecedent decision for an anaphor).

    The search space is represented as a nested list of mention pairs. The
    mention pairs are candidate arcs in the graph. The ith list contains all
    potential (mention, antecedent) pairs for the ith mention in the
    document. The antecedents are ordered by distance. For example,
    the third list contains the pairs (m_3, m_2), (m_3, m_1), (m_3, m_0),
    where m_j is the jth mention in the document.

    Args:
        doc (CoNLLDocument): The document to extract substructures from.

    Returns:
        (list(list((Mention, Mention)))): The nested list of mention pairs
        describing the search space for the substructures.
    """
    substructures = []

    # iterate over mentions
    for i, ana in enumerate(doc.system_mentions):
        for_anaphor_arcs = []

        # iterate in reversed order over candidate antecedents
        for ante in sorted(doc.system_mentions[:i], reverse=True):
            for_anaphor_arcs.append((ana, ante))

        substructures.append(for_anaphor_arcs)

    return substructures


class RankingPerceptron(perceptrons.Perceptron):
    """ A perceptron for mention ranking with latent antecedents. """
    def argmax(self, substructure, arc_information):
        """ Decoder for mention ranking with latent antecedents.

        Compute highest-scoring antecedent and highest-scoring antecedent
        consistent with gold coreference information for one anaphor.

        Args:
            substructure (list((Mention, Mention))): The list of mention pairs
                which define the search space for one substructure. For mention
                ranking, this list consists of all potential anaphor-antecedent
                pairs for one fixed anaphor in descending order, such as
                    (m_3, m_2), (m_2, m_1), (m_2, m_0)
            arc_information (dict((Mention, Mention),
                                  ((array, array, array), list(int), bool)):
                A mapping of arcs (= mention pairs) to information about these
                arcs. The information consists of the features, the costs for
                the arc (for each label), and whether predicting the arc to be
                coreferent is consistent with the gold annotation). The features
                are divided in three arrays: the first array contains the non-
                numeric features, the second array the numeric features, and the
                third array the values for the numeric features. The features
                are represented as integers via feature hashing.

        Returns:
            A 7-tuple describing the highest-scoring anaphor-antecedent
            decision, and the highest-scoring anaphor-antecedent decision
            consistent with the gold annotation. The tuple consists of:

                - **best_arcs** (*list((Mention, Mention))*): the
                  highest-scoring antecedent decision (the list contains only
                  one arc),
                - **best_labels** (*list(str)*): empty, the ranking approach
                  does not employ any labels,
                - **best_scores** (*list(float)*): the score of the
                  highest-scoring antecedent decision,
                - **best_cons_arcs** (*list((Mention, Mention))*): the
                  highest-scoring antecedent decision consistent with the gold
                  annotation (the list contains only one arc),
                - **best_cons_labels** (*list(str)*): empty, the ranking
                  approach does not employ any labels
                - **best_cons_scores** (*list(float)*): the score of the
                  highest-scoring antecedent decision consistent with the
                  gold information
                - **is_consistent** (*bool*): whether the highest-scoring
                  antecedent decision is consistent with the gold information.
        """
        best, max_val, best_cons, max_cons, best_is_consistent = \
            self.find_best_arcs(substructure, arc_information)

        return (
            [best],
            [],
            [max_val],
            [best_cons],
            [],
            [max_cons],
            best_is_consistent
        )


class RankingPerceptronClosest(perceptrons.Perceptron):
    """ A perceptron for mention ranking with closest antecedents for training.
    """
    def argmax(self, substructure, arc_information):
        """ Decoder for mention ranking with closest antecedents for training.

        Compute highest-scoring antecedent and closest gold antecedent for one
        anaphor.

        Args:
            substructure (list((Mention, Mention))): The list of mention pairs
                which define the search space for one substructure. For mention
                ranking, this list consists of all potential anaphor-antecedent
                pairs for one fixed anaphor in descending order, such as
                    (m_3, m_2), (m_3, m_1), (m_3, m_0)
            arc_information (dict((Mention, Mention),
                                  ((array, array, array), list(int), bool)):
                A mapping of arcs (= mention pairs) to information about these
                arcs. The information consists of the features, the costs for
                the arc (for each label), and whether predicting the arc to be
                coreferent is consistent with the gold annotation). The features
                are divided in three arrays: the first array contains the non-
                numeric features, the second array the numeric features, and the
                third array the values for the numeric features. The features
                are represented as integers via feature hashing.

        Returns:
            A 7-tuple describing the highest-scoring anaphor-antecedent
            decision, and the anaphor-antecedent pair with the closest gold
            antecedent. The tuple consists of:

                - **best_arcs** (*list((Mention, Mention))*): the
                  highest-scoring antecedent decision (the list contains only
                  one arc),
                - **best_labels** (*list(str)*): empty, the ranking approach
                  does not employ any labels,
                - **best_scores** (*list(float)*): the score of the
                  highest-scoring antecedent decision,
                - **best_cons_arcs** (*list((Mention, Mention))*): the
                  anaphor-antecedent pair with the closest gold antecedent (the
                  list contains only one arc),
                - **best_cons_labels** (*list(str)*): empty, the ranking
                  approach does not employ any labels
                - **best_cons_scores** (*list(float)*): the score of the
                  anaphor-antecedent pair with the closest gold antecedent
                - **is_consistent** (*bool*): whether the highest-scoring
                  antecedent decision is consistent with the gold information.
        """
        max_val = float("-inf")
        best = None

        max_cons = float("-inf")
        best_cons = None

        best_is_consistent = False

        for arc in substructure:
            score = self.score_arc(arc, arc_information)
            consistent = arc_information[arc][2]

            if score > max_val:
                best = arc
                max_val = score
                best_is_consistent = consistent

            # take closest
            if not best_cons and consistent:
                best_cons = arc
                max_cons = score

        return (
            [best],
            [],
            [max_val],
            [best_cons],
            [],
            [max_cons],
            best_is_consistent
        )


""" Manage learning from training data and making predictions on test data. """


import logging


__author__ = 'smartschat'


def learn(training_corpus, instance_extractor, perceptron):
    """ Learn a model for coreference resolution from training data.

    In particular, apply an instance/feature extractor to a training corpus and
    employ a machine learning model to learn a weight vector from these
    instances.

    Args:
        training_corpus (Corpus): The corpus to learn from.
        instance_extractor (InstanceExtracor): The instance extractor that
            defines the features and the structure of instances that are
            extracted during training.
        perceptron (Perceptron): A perceptron (including a decoder) that
            learns from the instances extracted by ``instance_extractor``.

    Returns:
        A tuple consisting of
            - **priors** (*dict(str,float)*): A prior weight for each label
              in the graphs representing the instances,
            - **weights** (*dict(str, array)*): A mapping of labels to weight
              vectors. For each label ``l``, ``weights[l]`` contains weights
              for each feature seen during training (for representing the
              features we employ *feature hashing*). If the graphs employed are
              not labeled, ``l`` is set to "+".
    """
    logging.info("Learning.")

    logging.info("\tExtracting instances and features.")
    substructures, arc_information = instance_extractor.extract(
        training_corpus)

    logging.info("\tFitting model parameters.")

    perceptron.fit(substructures, arc_information)

    return perceptron.get_model()


def predict(testing_corpus,
            instance_extractor,
            perceptron,
            coref_extractor):
    """ According to a learned model, predict coreference information.

    Args:
        testing_corpus (Corpus): The corpus to predict coreference on.
        instance_extractor (InstanceExtracor): The instance extracor that
            defines the features and the structure of instances that are
            extracted during testing.
        perceptron (Perceptron): A perceptron learned from training data.
        argmax_function (function): A decoder that computes the best-scoring
            coreference structure over a set of structures.
        coref_extractor (function): An extractor for consolidating pairwise
            predictions into coreference clusters.

    Returns:
        A tuple containing two dicts. The components are

            - **mention_entity_mapping** (*dict(Mention, int)*): A mapping of
              mentions to entity identifiers.
            - **antecedent_mapping** (*dict(Mention, Mention)*): A mapping of
              mentions to their antecedent (as determined by the
              ``coref_extractor``).
    """
    logging.info("Predicting.")

    logging.info("\tRemoving coreference annotations from corpus.")
    for doc in testing_corpus:
        doc.antecedent_decisions = {}
        for mention in doc.system_mentions:
            mention.attributes["antecedent"] = None
            mention.attributes["set_id"] = None

    logging.info("\tExtracting instances and features.")
    substructures, arc_information = instance_extractor.extract(testing_corpus)

    logging.info("\tDoing predictions.")
    arcs, labels, scores = perceptron.predict(substructures, arc_information)

    logging.info("\tClustering results.")

    return coref_extractor(arcs, labels, scores, perceptron.get_coref_labels())



""" Mixins. """


__author__ = 'smartschat'


class ComparableMixin:
    """ A mixin for deducing comparison operators from __lt__. """
    def __eq__(self, other):
        if self is None and other is not None:
            return False
        elif self is not None and other is None:
            return False
        else:
            return not self < other and not other < self

    def __ne__(self, other):
        return self < other or other < self

    def __gt__(self, other):
        return other < self

    def __ge__(self, other):
        return not self < other

    def __le__(self, other):
        return not other < self



import unittest

import nltk

from cort.core import documents
from cort.core import mention_extractor
from cort.core import mentions
from cort.core import spans


__author__ = 'smartschat'


class TestMentionExtractor(unittest.TestCase):
    def setUp(self):
        self.real_example = """#begin document (bn/voa/02/voa_0220); part 000
bn/voa/02/voa_0220   0    0    Unidentified    JJ  (TOP(S(NP(NP*          -   -   -   -            *    -
bn/voa/02/voa_0220   0    1          gunmen   NNS              *)         -   -   -   -            *    -
bn/voa/02/voa_0220   0    2              in    IN           (PP*          -   -   -   -            *    -
bn/voa/02/voa_0220   0    3           north    JJ      (NP(ADJP*          -   -   -   -            *    -
bn/voa/02/voa_0220   0    4         western    JJ              *)         -   -   -   -            *    -
bn/voa/02/voa_0220   0    5        Colombia   NNP            *)))         -   -   -   -         (GPE)   -
bn/voa/02/voa_0220   0    6            have   VBP           (VP*        have  -   -   -            *    -
bn/voa/02/voa_0220   0    7       massacred   VBN           (VP*    massacre  -   -   -            *    -
bn/voa/02/voa_0220   0    8              at    IN   (NP(QP(ADVP*          -   -   -   -   (CARDINAL*    -
bn/voa/02/voa_0220   0    9           least   JJS              *)         -   -   -   -            *    -
bn/voa/02/voa_0220   0   10          twelve    CD              *)         -   -   -   -            *)   -
bn/voa/02/voa_0220   0   11        peasants   NNS              *)         -   -   -   -            *    -
bn/voa/02/voa_0220   0   12              in    IN           (PP*          -   -   -   -            *    -
bn/voa/02/voa_0220   0   13             the    DT        (NP(NP*          -   -   -   -            *   (0
bn/voa/02/voa_0220   0   14          second    JJ              *          -   -   -   -     (ORDINAL)   -
bn/voa/02/voa_0220   0   15            such    JJ              *          -   -   -   -            *    -
bn/voa/02/voa_0220   0   16        incident    NN              *)   incident  -   2   -            *    -
bn/voa/02/voa_0220   0   17              in    IN           (PP*          -   -   -   -            *    -
bn/voa/02/voa_0220   0   18              as    RB        (NP(QP*          -   -   -   -       (DATE*    -
bn/voa/02/voa_0220   0   19            many    JJ              *)         -   -   -   -            *    -
bn/voa/02/voa_0220   0   20            days   NNS         *))))))        day  -   4   -            *)   0)
bn/voa/02/voa_0220   0   21               .     .             *))         -   -   -   -            *    -

bn/voa/02/voa_0220   0    0          Local    JJ    (TOP(S(NP*          -    -   -   -   *   (ARG0*             *    -
bn/voa/02/voa_0220   0    1         police   NNS             *)     police   -   -   -   *        *)            *    -
bn/voa/02/voa_0220   0    2            say   VBP          (VP*         say  01   1   -   *      (V*)            *    -
bn/voa/02/voa_0220   0    3             it   PRP   (SBAR(S(NP*)         -    -   -   -   *   (ARG1*        (ARG1*)   -
bn/voa/02/voa_0220   0    4             's   VBZ          (VP*          be  01   1   -   *        *           (V*)   -
bn/voa/02/voa_0220   0    5            not    RB             *          -    -   -   -   *        *    (ARGM-NEG*)   -
bn/voa/02/voa_0220   0    6          clear    JJ        (ADJP*)         -    -   -   -   *        *        (ARG2*)   -
bn/voa/02/voa_0220   0    7            who    WP   (SBAR(WHNP*)         -    -   -   -   *        *             *    -
bn/voa/02/voa_0220   0    8            was   VBD        (S(VP*          be   -   1   -   *        *             *    -
bn/voa/02/voa_0220   0    9    responsible    JJ        (ADJP*          -    -   -   -   *        *             *    -
bn/voa/02/voa_0220   0   10            for    IN          (PP*          -    -   -   -   *        *             *    -
bn/voa/02/voa_0220   0   11            the    DT          (NP*          -    -   -   -   *        *             *   (0
bn/voa/02/voa_0220   0   12       massacre    NN    *))))))))))   massacre   -   -   -   *        *)            *    0)
bn/voa/02/voa_0220   0   13              .     .            *))         -    -   -   -   *        *             *    -

#end document
"""

        self.another_real_example = """#begin document (mz/sinorama/10/ectb_1050); part 006
mz/sinorama/10/ectb_1050        6       0       What    WP      (TOP(SBARQ(WHNP*)       -       -       -       -       *       (R-ARG1*)       -
mz/sinorama/10/ectb_1050        6       1       does    VBZ     (SQ*    do      -       7       -       *       *       -
mz/sinorama/10/ectb_1050        6       2       this    DT      (NP*)   -       -       -       -       *       (ARG0*) -
mz/sinorama/10/ectb_1050        6       3       tell    VB      (VP*    tell    01      1       -       *       (V*)    -
mz/sinorama/10/ectb_1050        6       4       us      PRP     (NP*)   -       -       -       -       *       (ARG2*) -
mz/sinorama/10/ectb_1050        6       5       about   IN      (PP*    -       -       -       -       *       (ARG1*  -
mz/sinorama/10/ectb_1050        6       6       the     DT      (NP(NP* -       -       -       -       *       *       -
mz/sinorama/10/ectb_1050        6       7       transformation  NN      *)      transformation  -       1       -       *       *       -
mz/sinorama/10/ectb_1050        6       8       of      IN      (PP*    -       -       -       -       *       *       -
mz/sinorama/10/ectb_1050        6       9       Taiwan  NNP     (NP(NP* -       -       -       -       (GPE)   *       -
mz/sinorama/10/ectb_1050        6       10      's      POS     *)      -       -       -       -       *       *       -
mz/sinorama/10/ectb_1050        6       11      townships       NNS     *)))))) township        -       1       -       *       *)      -
mz/sinorama/10/ectb_1050        6       12      ?       .       *))     -       -       -       -       *       *       -

#end	document"""

        self.real_document = documents.CoNLLDocument(self.real_example)
        self.another_real_document = documents.CoNLLDocument(
            self.another_real_example)

        self.tree = nltk.ParentedTree.fromstring(
            "(NP (NP (NP (PRP$ his) (NN brother) (POS 's)) (NN wedding)) "
            "(PP (IN in) (NP (NNP Khan) (NNPS Younes))))")

        self.proper_name_mention_tree = nltk.ParentedTree.fromstring(
            "(NP (NNP Taiwan) (POS 's))")
        self.proper_name_mention_ner = ["GPE", "NONE"]

        self.apposition_tree = nltk.ParentedTree.fromstring(
            "(NP (NP (NP (NNP Secretary)) (PP (IN of) (NP (NNP State)))) "
            "(NP (NNP Madeleine) (NNP Albright)))")

        self.apposition_ner = ["NONE", "NONE", "NONE", "PERSON", "PERSON"]

        self.more_proper_name_tree = nltk.ParentedTree.fromstring(
            "(NP (NP (DT the) (NNP General) (NNP Secretary)) (PP (IN of) "
            "(NP (DT the) (NNP CCP))))")

        self.more_proper_name_ner = ["NONE", "NONE", "NONE", "NONE", "NONE",
                                     "ORG"]

    def test_extract_system_mentions(self):
        expected_spans = sorted([
            spans.Span(0, 1),
            spans.Span(0, 5),
            spans.Span(3, 5),
            spans.Span(5, 5),
            spans.Span(8, 10),
            spans.Span(8, 11),
            spans.Span(13, 16),
            spans.Span(13, 20),
            spans.Span(14, 14),
            spans.Span(18, 20),
            spans.Span(22, 23),
            spans.Span(25, 25),
            spans.Span(33, 34)
        ])

        self.assertEqual(expected_spans,
                         [mention.span for
                          mention in mention_extractor.extract_system_mentions(
                             self.real_document, filter_mentions=False)[1:]])

        expected_spans = sorted([
            spans.Span(2, 2),
            spans.Span(4, 4),
            spans.Span(6, 7),
            spans.Span(6, 11),
            spans.Span(9, 10),
            spans.Span(9, 11)
        ])

        self.assertEqual(expected_spans,
                         [mention.span for
                          mention in mention_extractor.extract_system_mentions(
                             self.another_real_document,
                             filter_mentions=False)[1:]])

        expected_spans = sorted([
            spans.Span(2, 2),
            spans.Span(4, 4),
            spans.Span(6, 11),
            spans.Span(9, 10),
            spans.Span(9, 11)
        ])

        self.assertEqual(expected_spans,
                         [mention.span for
                          mention in mention_extractor.extract_system_mentions(
                             self.another_real_document,
                             filter_mentions=True)[1:]])

    def test_post_process_same_head_largest_span(self):
        all_mentions = {
            mentions.Mention(
                None,
                spans.Span(0, 3),
                {"tokens": [], "type": "NOM", "head_index": 0,
                 "head_span": spans.Span(3, 3)}),
            mentions.Mention(
                None,
                spans.Span(0, 6),
                {"tokens": [], "type": "NOM", "head_index": 0,
                 "head_span": spans.Span(3, 3)}),
            mentions.Mention(
                None,
                spans.Span(0, 2),
                {"tokens": [], "type": "NOM", "head_index": 0,
                 "head_span": spans.Span(1, 1)}),
            mentions.Mention(
                None,
                spans.Span(5, 6),
                {"tokens": [], "type": "NOM", "head_index": 0,
                 "head_span": spans.Span(5, 6)}),
            mentions.Mention(
                None,
                spans.Span(0, 0),
                {"tokens": [], "type": "NOM", "head_index": 0,
                 "head_span": spans.Span(0, 0)})}

        expected_mentions = sorted([
            mentions.Mention(
                None,
                spans.Span(0, 6),
                {"tokens": [], "type": "NOM", "head_index": 0,
                 "head_span": spans.Span(3, 3)}),
            mentions.Mention(
                None,
                spans.Span(0, 2),
                {"tokens": [], "type": "NOM", "head_index": 0,
                 "head_span": spans.Span(1, 1)}),
            mentions.Mention(
                None,
                spans.Span(5, 6),
                {"tokens": [], "type": "NOM", "head_index": 0,
                 "head_span": spans.Span(5, 6)}),
            mentions.Mention(
                None,
                spans.Span(0, 0),
                {"tokens": [], "type": "NOM", "head_index": 0,
                 "head_span": spans.Span(0, 0)})
        ])

        self.assertEqual(
            expected_mentions,
            mention_extractor.post_process_same_head_largest_span(all_mentions))

        all_mentions_2 = {
            mentions.Mention(
                None,
                spans.Span(0, 1),
                {"tokens": ["Taiwan", "'s"], "type": "NAM", "head_index": 0,
                 "head_span": spans.Span(0, 0)}),
            mentions.Mention(
                None,
                spans.Span(0, 0),
                {"tokens": ["Taiwan"], "type": "NAM", "head_index": 0,
                 "head_span": spans.Span(0, 0)}),
            mentions.Mention(
                None,
                spans.Span(2, 3),
                {"tokens": ["the", "CCP"], "type": "NAM", "head_index": 1,
                 "head_span": spans.Span(3, 3)}),
            mentions.Mention(
                None,
                spans.Span(3, 3),
                {"tokens": ["CCP"], "type": "NAM", "head_index": 0,
                 "head_span": spans.Span(3, 3)})}

        expected_mentions_2 = sorted([
            mentions.Mention(
                None,
                spans.Span(0, 1),
                {"tokens": ["Taiwan", "'s"], "type": "NAM", "head_index": 0,
                 "head_span": spans.Span(0, 0)}),
            mentions.Mention(
                None,
                spans.Span(2, 3),
                {"tokens": ["the", "CCP"], "type": "NAM", "head_index": 1,
                 "head_span": spans.Span(3, 3)}),
        ])

        self.assertEqual(
            expected_mentions_2,
            mention_extractor.post_process_same_head_largest_span(
                all_mentions_2))

    def test_post_process_embedded_head_largest_span(self):
        all_mentions_1 = {
            mentions.Mention(
                None,
                spans.Span(0, 3),
                {"tokens": [], "type": "NOM", "head_index": 0,
                 "head_span": spans.Span(3, 3)}),
            mentions.Mention(
                None,
                spans.Span(0, 6),
                {"tokens": [], "type": "NOM", "head_index": 0,
                 "head_span": spans.Span(2, 3)}),
            mentions.Mention(
                None,
                spans.Span(0, 2),
                {"tokens": [], "type": "NOM", "head_index": 0,
                 "head_span": spans.Span(1, 1)}),
            mentions.Mention(
                None,
                spans.Span(5, 6),
                {"tokens": [], "type": "NOM", "head_index": 0,
                 "head_span": spans.Span(5, 6)})
        }

        expected_mentions_1 = sorted([
            mentions.Mention(
                None,
                spans.Span(0, 6),
                {"tokens": [], "type": "NOM", "head_index": 0,
                 "head_span": spans.Span(2, 3)}),
            mentions.Mention(
                None,
                spans.Span(0, 2),
                {"tokens": [], "type": "NOM", "head_index": 0,
                 "head_span": spans.Span(1, 1)}),
            mentions.Mention(
                None,
                spans.Span(5, 6),
                {"tokens": [], "type": "NOM", "head_index": 0,
                 "head_span": spans.Span(5, 6)})
        ])

        self.assertEqual(
            expected_mentions_1,
            mention_extractor.post_process_embedded_head_largest_span(
                all_mentions_1))

    def test_post_process_appositions(self):
        three_children_tree = nltk.ParentedTree.fromstring(
            "(NP (NP (NP (NP (DT The) (NNP ROC) (POS 's)) (NN ambassador)) "
            "(PP (IN to) (NP (NNP Nicaragua)))) (, ,) (NP (NNP Antonio) "
            "(NNP Tsai)) (, ,))")

        three_children_all_mentions = {
            mentions.Mention(
                None,
                spans.Span(0, 6),
                {"tokens": ["The", "ROC", "'s", "ambassador", "to",
                            "Nicaragua", ",", "Antonio", "Tsai"],
                 "is_apposition": True, "type": "NAM",
                 "parse_tree": three_children_tree}),
            mentions.Mention(
                None,
                spans.Span(0, 4),
                {"tokens": ["The", "ROC", "'s", "ambassador", "to",
                            "Nicaragua"],
                 "is_apposition": False, "type": "NOM",
                 "parse_tree": three_children_tree[0]}),
            mentions.Mention(
                None,
                spans.Span(0, 3),
                {"tokens": ["The", "ROC", "'s", "ambassador"],
                 "is_apposition": False, "type": "NOM",
                 "parse_tree": three_children_tree[0][0]}),
            mentions.Mention(
                None,
                spans.Span(0, 2),
                {"tokens": ["The", "ROC", "'s"], "is_apposition": False,
                 "type": "NAM", "parse_tree": three_children_tree[0][0][0]}),
            mentions.Mention(
                None,
                spans.Span(4, 4),
                {"tokens": ["Nicaragua"], "is_apposition": False,
                 "type": "NAM", "parse_tree": three_children_tree[0][1][1]}),
            mentions.Mention(
                None,
                spans.Span(5, 6),
                {"tokens": ["Antonio", "Tsai"], "is_apposition": False,
                 "type": "NAM", "parse_tree": three_children_tree[2]})}

        three_children_expected = sorted([
            mentions.Mention(
                None,
                spans.Span(0, 6),
                {"tokens": ["The", "ROC", "'s", "ambassador", "to",
                            "Nicaragua", ",", "Antonio", "Tsai"],
                 "is_apposition": True, "type": "NAM",
                 "parse_tree": three_children_tree}),
            mentions.Mention(
                None,
                spans.Span(0, 3),
                {"tokens": ["The", "ROC", "'s", "ambassador"],
                 "is_apposition": False, "type": "NOM",
                 "parse_tree": three_children_tree[0][0]}),
            mentions.Mention(
                None,
                spans.Span(0, 2),
                {"tokens": ["The", "ROC", "'s"], "is_apposition": False,
                 "type": "NAM", "parse_tree": three_children_tree[0][0][0]}),
            mentions.Mention(
                None,
                spans.Span(4, 4),
                {"tokens": ["Nicaragua"], "is_apposition": False,
                 "type": "NAM", "parse_tree": three_children_tree[0][1][1]}),
        ])

        self.assertEqual(
            three_children_expected,
            mention_extractor.post_process_appositions(
                three_children_all_mentions))

        two_children_tree = nltk.ParentedTree.fromstring(
            "(NP (NP (NP (NNP Secretary)) (PP (IN of) (NP (NNP State)))) "
            "(NP (NNP Madeleine) (NNP Albright)))")

        two_children_all_mentions = {
            mentions.Mention(
                None,
                spans.Span(0, 4),
                {"tokens": ["Secretary", "of", "Sate", "Madeleine",
                            "Albright"],
                 "is_apposition": True, "type": "NAM",
                 "parse_tree": two_children_tree}),
            mentions.Mention(
                None,
                spans.Span(0, 0),
                {"tokens": ["Secretary"], "is_apposition": False,
                 "type": "NAM", "parse_tree": two_children_tree[0][0]}),
            mentions.Mention(
                None,
                spans.Span(0, 2),
                {"tokens": ["Secretary", "of", "State"],
                 "is_apposition": False, "type": "NAM",
                 "parse_tree": two_children_tree[0]}),
            mentions.Mention(
                None,
                spans.Span(2, 2),
                {"tokens": ["State"], "is_apposition": False,
                 "type": "NAM", "parse_tree": two_children_tree[0][1][1]}),
            mentions.Mention(
                None,
                spans.Span(2, 2),
                {"tokens": ["Madeleine", "Albright"], "is_apposition": False,
                 "type": "NAM", "parse_tree": two_children_tree[1]})}

        two_children_expected = sorted([
            mentions.Mention(
                None,
                spans.Span(0, 4),
                {"tokens": ["Secretary", "of", "Sate", "Madeleine",
                            "Albright"],
                 "is_apposition": True, "type": "NAM",
                 "parse_tree": two_children_tree})
        ])

        self.assertEqual(
            two_children_expected,
            mention_extractor.post_process_appositions(
                two_children_all_mentions))

if __name__ == '__main__':
    unittest.main()



import unittest

from cort.core.mentions import Mention
from cort.core.spans import Span
from cort.core.documents import CoNLLDocument

import nltk

__author__ = 'smartschat'


class TestDocuments(unittest.TestCase):
    def setUp(self):
        self.real_example = """#begin document (bn/voa/02/voa_0220); part 000
bn/voa/02/voa_0220   0    0    Unidentified    JJ  (TOP(S(NP(NP*          -   -   -   -            *    -
bn/voa/02/voa_0220   0    1          gunmen   NNS              *)         -   -   -   -            *    -
bn/voa/02/voa_0220   0    2              in    IN           (PP*          -   -   -   -            *    -
bn/voa/02/voa_0220   0    3           north    JJ      (NP(ADJP*          -   -   -   -            *    -
bn/voa/02/voa_0220   0    4         western    JJ              *)         -   -   -   -            *    -
bn/voa/02/voa_0220   0    5        Colombia   NNP            *)))         -   -   -   -         (GPE)   -
bn/voa/02/voa_0220   0    6            have   VBP           (VP*        have  -   -   -            *    -
bn/voa/02/voa_0220   0    7       massacred   VBN           (VP*    massacre  -   -   -            *    -
bn/voa/02/voa_0220   0    8              at    IN   (NP(QP(ADVP*          -   -   -   -   (CARDINAL*    -
bn/voa/02/voa_0220   0    9           least   JJS              *)         -   -   -   -            *    -
bn/voa/02/voa_0220   0   10          twelve    CD              *)         -   -   -   -            *)   -
bn/voa/02/voa_0220   0   11        peasants   NNS              *)         -   -   -   -            *    -
bn/voa/02/voa_0220   0   12              in    IN           (PP*          -   -   -   -            *    -
bn/voa/02/voa_0220   0   13             the    DT        (NP(NP*          -   -   -   -            *   (0
bn/voa/02/voa_0220   0   14          second    JJ              *          -   -   -   -     (ORDINAL)   -
bn/voa/02/voa_0220   0   15            such    JJ              *          -   -   -   -            *    -
bn/voa/02/voa_0220   0   16        incident    NN              *)   incident  -   2   -            *    -
bn/voa/02/voa_0220   0   17              in    IN           (PP*          -   -   -   -            *    -
bn/voa/02/voa_0220   0   18              as    RB        (NP(QP*          -   -   -   -       (DATE*    -
bn/voa/02/voa_0220   0   19            many    JJ              *)         -   -   -   -            *    -
bn/voa/02/voa_0220   0   20            days   NNS         *))))))        day  -   4   -            *)   0)
bn/voa/02/voa_0220   0   21               .     .             *))         -   -   -   -            *    -

bn/voa/02/voa_0220   0    0          Local    JJ    (TOP(S(NP*          -    -   -   -   *   (ARG0*             *    -
bn/voa/02/voa_0220   0    1         police   NNS             *)     police   -   -   -   *        *)            *    -
bn/voa/02/voa_0220   0    2            say   VBP          (VP*         say  01   1   -   *      (V*)            *    -
bn/voa/02/voa_0220   0    3             it   PRP   (SBAR(S(NP*)         -    -   -   -   *   (ARG1*        (ARG1*)   -
bn/voa/02/voa_0220   0    4             's   VBZ          (VP*          be  01   1   -   *        *           (V*)   -
bn/voa/02/voa_0220   0    5            not    RB             *          -    -   -   -   *        *    (ARGM-NEG*)   -
bn/voa/02/voa_0220   0    6          clear    JJ        (ADJP*)         -    -   -   -   *        *        (ARG2*)   -
bn/voa/02/voa_0220   0    7            who    WP   (SBAR(WHNP*)         -    -   -   -   *        *             *    -
bn/voa/02/voa_0220   0    8            was   VBD        (S(VP*          be   -   1   -   *        *             *    -
bn/voa/02/voa_0220   0    9    responsible    JJ        (ADJP*          -    -   -   -   *        *             *    -
bn/voa/02/voa_0220   0   10            for    IN          (PP*          -    -   -   -   *        *             *    -
bn/voa/02/voa_0220   0   11            the    DT          (NP*          -    -   -   -   *        *             *   (0
bn/voa/02/voa_0220   0   12       massacre    NN    *))))))))))   massacre   -   -   -   *        *)            *    0)
bn/voa/02/voa_0220   0   13              .     .            *))         -    -   -   -   *        *             *    -

#end document
"""
        self.complicated_mention_example = """#begin document (/test2); part 000
test2	0	0	This    NN   (NP*	-   -   -   -   -   (0)
test2	0	1	is  NN	*   -   -   -   -   -   -
test2	0	2	just    NN   *	-   -   -   -   -   -
test2	0	3	a   NN   *	-   -   -   -   -   (0|(1)
test2	0	4	test    NN   *	-   -   -   -   -   0)
test2	0	5	.   NN   *)	-   -   -   -   -   -

test2	0	0	It  NN   (NP*	-   -   -   -   -   (1)|(4
test2	0	1	shows   NN   *	-   -   -   -   -   -
test2	0	2	that    NN   *	-   -   -   -   -   (2)
test2	0	3	the NN   *	-   -   -   -   -   (2|(0
test2	0	4	scorer  NN   *	-   -   -   -   -   2)|4)
test2	0	5	works   NN   *	-   -   -   -   -   0)
test2	0	6	.   NN   *)	-   -   -   -   -   -

#end document
"""

        self.another_real_example = """#begin document (mz/sinorama/10/ectb_1050); part 006
mz/sinorama/10/ectb_1050        6       0       What    WP      (TOP(SBARQ(WHNP*)       -       -       -       -       *       (R-ARG1*)       -
mz/sinorama/10/ectb_1050        6       1       does    VBZ     (SQ*    do      -       7       -       *       *       -
mz/sinorama/10/ectb_1050        6       2       this    DT      (NP*)   -       -       -       -       *       (ARG0*) -
mz/sinorama/10/ectb_1050        6       3       tell    VB      (VP*    tell    01      1       -       *       (V*)    -
mz/sinorama/10/ectb_1050        6       4       us      PRP     (NP*)   -       -       -       -       *       (ARG2*) -
mz/sinorama/10/ectb_1050        6       5       about   IN      (PP*    -       -       -       -       *       (ARG1*  -
mz/sinorama/10/ectb_1050        6       6       the     DT      (NP(NP* -       -       -       -       *       *       -
mz/sinorama/10/ectb_1050        6       7       transformation  NN      *)      transformation  -       1       -       *       *       -
mz/sinorama/10/ectb_1050        6       8       of      IN      (PP*    -       -       -       -       *       *       -
mz/sinorama/10/ectb_1050        6       9       Taiwan  NNP     (NP(NP* -       -       -       -       (GPE)   *       -
mz/sinorama/10/ectb_1050        6       10      's      POS     *)      -       -       -       -       *       *       -
mz/sinorama/10/ectb_1050        6       11      townships       NNS     *)))))) township        -       1       -       *       *)      -
mz/sinorama/10/ectb_1050        6       12      ?       .       *))     -       -       -       -       *       *       -

#end	document
"""

        self.yemen_example = """#begin document (bn/abc/00/abc_0030); part 000
bn/abc/00/abc_0030      0       0       Intelligence    NN      (TOP(S(NP*      -       -       -       -       *       (ARG0*  *       -
bn/abc/00/abc_0030      0       1       sources NNS     *)      source  -       3       -       *       *)      *       -
bn/abc/00/abc_0030      0       2       say     VBP     (VP*    say     01      1       -       *       (V*)    *       -
bn/abc/00/abc_0030      0       3       the     DT      (SBAR(S(NP*     -       -       -       -       *       (ARG1*  (ARG1*  -
bn/abc/00/abc_0030      0       4       target  NN      *)      target  -       2       -       *       *       *)      -
bn/abc/00/abc_0030      0       5       was     VBD     (VP*    be      -       1       -       *       *       *       -
bn/abc/00/abc_0030      0       6       to      TO      (S(VP*  -       -       -       -       *       *       *       -
bn/abc/00/abc_0030      0       7       be      VB      (VP*    be      01      1       -       *       *       (V*)    -
bn/abc/00/abc_0030      0       8       a       DT      (NP(NP* -       -       -       -       *       *       *       -
bn/abc/00/abc_0030      0       9       destroyer       NN      *)      -       -       -       -       *       *       *       -
bn/abc/00/abc_0030      0       10      ,       ,       *       -       -       -       -       *       *       *       -
bn/abc/00/abc_0030      0       11      the     DT      (NP*    -       -       -       -       *       *       *       -
bn/abc/00/abc_0030      0       12      ``      ``      *       -       -       -       -       *       *       *       -
bn/abc/00/abc_0030      0       13      USS     NNP     (NP*    -       -       -       -       (PRODUCT*       *       *       -
bn/abc/00/abc_0030      0       14      The     NNP     *       -       -       -       -       *       *       *       -
bn/abc/00/abc_0030      0       15      Sullivans       NNP     *)      -       -       -       -       *)      *       *       -
bn/abc/00/abc_0030      0       16      ,       ,       *       -       -       -       -       *       *       *       -
bn/abc/00/abc_0030      0       17      ''      ''      *       -       -       -       -       *       *       *       -
bn/abc/00/abc_0030      0       18      which   WDT     (SBAR(WHNP*)    -       -       -       -       *       *       *       -
bn/abc/00/abc_0030      0       19      refueled        VBD     (S(VP*  refuel  -       -       -       *       *       *       -
bn/abc/00/abc_0030      0       20      in      IN      (PP*    -       -       -       -       *       *       *       -
bn/abc/00/abc_0030      0       21      Yemen   NNP     (NP(NP(NP*      -       -       -       -       (GPE)   *       *       -
bn/abc/00/abc_0030      0       22      's      POS     *)      -       -       -       -       *       *       *       -
bn/abc/00/abc_0030      0       23      port    NN      *)      port    -       1       -       (LOC*   *       *       -
bn/abc/00/abc_0030      0       24      of      IN      (PP*    -       -       -       -       *       *       *       -
bn/abc/00/abc_0030      0       25      Aden    NNP     (NP*))))        -       -       -       -       *)      *       *       -
bn/abc/00/abc_0030      0       26      in      IN      (PP*    -       -       -       -       *       *       *       -
bn/abc/00/abc_0030      0       27      January NNP     (NP*))))))))))))))      -       -       -       -       (DATE)  *)      *       -
bn/abc/00/abc_0030      0       28      .       .       *))     -       -       -       -       *       *       *       -

#end document
"""

        self.real_document = CoNLLDocument(self.real_example)
        self.complicated_mention_document = CoNLLDocument(
            self.complicated_mention_example)
        self.another_real_document = CoNLLDocument(self.another_real_example)
        self.yemen_document = CoNLLDocument(self.yemen_example)

        self.maxDiff = None

    def test_get_identifier(self):
        self.assertEqual("(bn/voa/02/voa_0220); part 000",
                         self.real_document.identifier)

    def test_get_tokens(self):
        tokens = ["Unidentified", "gunmen", "in", "north", "western",
                  "Colombia", "have", "massacred", "at", "least", "twelve",
                  "peasants", "in", "the", "second", "such", "incident", "in",
                  "as", "many", "days", ".", "Local", "police", "say",
                  "it", "'s", "not", "clear", "who", "was", "responsible",
                  "for", "the", "massacre", "."]
        self.assertEqual(tokens, self.real_document.tokens)

    def test_get_ner(self):
        ner = ["NONE"] * 36
        ner[5:6] = ["GPE"]
        ner[8:11] = ["CARDINAL"] * 3
        ner[14:15] = ["ORDINAL"]
        ner[18:21] = ["DATE"] * 3

        self.assertEqual(ner, self.real_document.ner)

    def test_get_coref(self):
        simple = {
            Span(13, 20): 0,
            Span(33, 34): 0
        }

        complicated = {
            Span(0, 0): 0,
            Span(3, 3): 1,
            Span(3, 4): 0,
            Span(6, 6): 1,
            Span(6, 10): 4,
            Span(8, 8): 2,
            Span(9, 10): 2,
            Span(9, 11): 0
        }

        self.assertEqual(simple, self.real_document.coref)
        self.assertEqual(complicated, self.complicated_mention_document.coref)

    def test_extract_sentence_spans(self):
        sentence_spans = [Span(0, 21), Span(22, 35)]

        self.assertEqual(sentence_spans,
                         self.real_document.sentence_spans)

    def test_get_sentence_id_and_span(self):
        expected = 1, Span(22, 35)
        self.assertEqual(expected, self.real_document.get_sentence_id_and_span(
            Span(23, 24)))

    def test_parse(self):
        expected = nltk.ParentedTree.fromstring(
            "(TOP (S (NP (JJ Local) (NNS police)) (VP (VBP say) "
            "(SBAR (S (NP (PRP it)) (VP (VBZ 's) (RB not) "
            "(ADJP (JJ clear)) (SBAR (WHNP (WP who)) (S (VP (VBD was) "
            "(ADJP (JJ responsible) (PP (IN for) (NP (DT the) (NN "
            "massacre))))))))))) (. .)))")
        self.assertEqual(expected, self.real_document.parse[1])

    def test_get_string_representation(self):
        expected = """#begin document (/test2); part 000
test2	0	0	This	NN	(NP*	-	-	-	-	-	(0|(1)
test2	0	1	is	NN	*	-	-	-	-	-	0)
test2	0	2	just	NN	*	-	-	-	-	-	-
test2	0	3	a	NN	*	-	-	-	-	-	-
test2	0	4	test	NN	*	-	-	-	-	-	(1
test2	0	5	.	NN	*)	-	-	-	-	-	(2|1)

test2	0	0	It	NN	(NP*	-	-	-	-	-	2)
test2	0	1	shows	NN	*	-	-	-	-	-	(3)
test2	0	2	that	NN	*	-	-	-	-	-	(3)
test2	0	3	the	NN	*	-	-	-	-	-	-
test2	0	4	scorer	NN	*	-	-	-	-	-	-
test2	0	5	works	NN	*	-	-	-	-	-	-
test2	0	6	.	NN	*)	-	-	-	-	-	-
#end document
"""

        self.complicated_mention_document.system_mentions = [
            Mention(self.complicated_mention_document, Span(0, 0),
                    {"set_id": 1}),
            Mention(self.complicated_mention_document, Span(0, 1),
                    {"set_id": 0}),
            Mention(self.complicated_mention_document, Span(4, 5),
                    {"set_id": 1}),
            Mention(self.complicated_mention_document, Span(5, 6),
                    {"set_id": 2}),
            Mention(self.complicated_mention_document, Span(7, 7),
                    {"set_id": 3}),
            Mention(self.complicated_mention_document, Span(8, 8),
                    {"set_id": 3}),
        ]

        self.assertEqual(
            expected,
            self.complicated_mention_document.get_string_representation())


if __name__ == '__main__':
    unittest.main()
